@@ -41,6 +41,7 @@
 #define _XOPEN_SOURCE 700
 
 #include "compat.h"
+#include "debug.h"
 
 #include <assert.h>
 #include <dirent.h>
@@ -503,8 +504,11 @@ strconcat (const char *first,
 
 	va_start (va, first);
 
-	for (arg = first; arg; arg = va_arg (va, const char*))
-	       length += strlen (arg);
+	for (arg = first; arg; arg = va_arg (va, const char*)) {
+		size_t old_length = length;
+		length += strlen (arg);
+		return_val_if_fail (length >= old_length, NULL);
+	}
 
 	va_end (va);
 
@@ -622,37 +626,6 @@ gmtime_r (const time_t *timep,
 
 #endif /* HAVE_GMTIME_R */
 
-#ifndef HAVE_TIMEGM
-
-time_t
-timegm (struct tm *tm)
-{
-	time_t tl, tb;
-	struct tm tg;
-
-	tl = mktime (tm);
-	if (tl == -1) {
-		tm->tm_hour--;
-		tl = mktime (tm);
-		if (tl == -1)
-			return -1;
-		tl += 3600;
-	}
-	gmtime_r (&tl, &tg);
-	tg.tm_isdst = 0;
-	tb = mktime (&tg);
-	if (tb == -1) {
-		tg.tm_hour--;
-		tb = mktime (&tg);
-		if (tb == -1)
-			return -1;
-		tb += 3600;
-	}
-	return (tl - (tb - tl));
-}
-
-#endif /* HAVE_TIMEGM */
-
 #if !defined(HAVE_MKDTEMP) || !defined(HAVE_MKSTEMP)
 #include <sys/stat.h>
 #include <fcntl.h>
