@@ -61,8 +61,10 @@ struct p11_kit_iter {
 
 	/* Iterator matching data */
 	CK_INFO match_module;
+	CK_SLOT_INFO match_slot;
 	CK_TOKEN_INFO match_token;
 	CK_ATTRIBUTE *match_attrs;
+	CK_SLOT_ID match_slot_id;
 	Callback *callbacks;
 
 	/* The input modules */
@@ -84,6 +86,7 @@ struct p11_kit_iter {
 	CK_SLOT_ID slot;
 	CK_SESSION_HANDLE session;
 	CK_OBJECT_HANDLE object;
+	CK_SLOT_INFO slot_info;
 	CK_TOKEN_INFO token_info;
 
 	/* And various flags */
@@ -103,7 +106,7 @@ struct p11_kit_iter {
  * @P11_KIT_ITER_WANT_WRITABLE: Try to open read-write sessions when
  *   iterating over obojects.
  *
- * Various flags controling the behavior of the iterator.
+ * Various flags controlling the behavior of the iterator.
  */
 
 /**
@@ -164,6 +167,7 @@ p11_kit_iter_set_uri (P11KitIter *iter,
 {
 	CK_ATTRIBUTE *attrs;
 	CK_TOKEN_INFO *tinfo;
+	CK_SLOT_INFO *sinfo;
 	CK_INFO *minfo;
 	CK_ULONG count;
 
@@ -178,19 +182,26 @@ p11_kit_iter_set_uri (P11KitIter *iter,
 			attrs = p11_kit_uri_get_attributes (uri, &count);
 			iter->match_attrs = p11_attrs_buildn (NULL, attrs, count);
 
+			iter->match_slot_id = p11_kit_uri_get_slot_id (uri);
+
 			minfo = p11_kit_uri_get_module_info (uri);
 			if (minfo != NULL)
 				memcpy (&iter->match_module, minfo, sizeof (CK_INFO));
 
+			sinfo = p11_kit_uri_get_slot_info (uri);
+			if (sinfo != NULL)
+				memcpy (&iter->match_slot, sinfo, sizeof (CK_SLOT_INFO));
+
 			tinfo = p11_kit_uri_get_token_info (uri);
 			if (tinfo != NULL)
 				memcpy (&iter->match_token, tinfo, sizeof (CK_TOKEN_INFO));
 		}
 	} else {
-		/* Match any module version number*/
+		/* Match any module version number and slot ID */
 		memset (&iter->match_module, 0, sizeof (iter->match_module));
 		iter->match_module.libraryVersion.major = (CK_BYTE)-1;
 		iter->match_module.libraryVersion.minor = (CK_BYTE)-1;
+		iter->match_slot_id = (CK_SLOT_ID)-1;
 	}
 }
 
@@ -506,6 +517,11 @@ move_next_session (P11KitIter *iter)
 		iter->slot = iter->slots[iter->saw_slots++];
 
 		assert (iter->module != NULL);
+		if (iter->match_slot_id != (CK_SLOT_ID)-1 && iter->slot != iter->match_slot_id)
+			continue;
+		rv = (iter->module->C_GetSlotInfo) (iter->slot, &iter->slot_info);
+		if (rv != CKR_OK || !p11_match_uri_slot_info (&iter->match_slot, &iter->slot_info))
+			continue;
 		rv = (iter->module->C_GetTokenInfo) (iter->slot, &iter->token_info);
 		if (rv != CKR_OK || !p11_match_uri_token_info (&iter->match_token, &iter->token_info))
 			continue;
@@ -674,13 +690,30 @@ p11_kit_iter_get_slot (P11KitIter *iter)
 	return iter->slot;
 }
 
+/**
+ * p11_kit_iter_get_slot_info:
+ * @iter: the iterator
+ *
+ * Get the slot info for the slot which the current matching object is on.
+ *
+ * This can only be called after p11_kit_iter_next() succeeds.
+ *
+ * Returns: the slot of the current matching object.
+ */
+CK_SLOT_INFO *
+p11_kit_iter_get_slot_info (P11KitIter *iter)
+{
+	return_val_if_fail (iter != NULL, NULL);
+	return &iter->slot_info;
+}
+
 /**
  * p11_kit_iter_get_token:
  * @iter: the iterator
  *
  * Get the token info for the token which the current matching object is on.
  *
- * This can only be called after p11_kit_iter_next(0 succeeds.
+ * This can only be called after p11_kit_iter_next() succeeds.
  *
  * Returns: the slot of the current matching object.
  */
@@ -732,9 +765,9 @@ p11_kit_iter_get_object (P11KitIter *iter)
 
 /**
  * p11_kit_iter_destroy_object:
- * @iter: teh iterator
+ * @iter: the iterator
  *
- * Destory the current matching object.
+ * Destroy the current matching object.
  *
  * This can only be called after p11_kit_iter_next() succeeds.
  *
