@@ -251,6 +251,8 @@ free_module_unlocked (void *data)
 		assert (mod->initialize_thread == 0);
 	}
 
+	p11_virtual_uninit (&mod->virt);
+
 	if (mod->loaded_destroy)
 		mod->loaded_destroy (mod->loaded_module);
 
@@ -518,15 +520,15 @@ take_config_and_load_module_inlock (char **name,
 		rv = load_module_from_file_inlock (*name, filename, &mod);
 		if (rv != CKR_OK)
 			goto out;
-
-		/*
-		 * We support setting of CK_C_INITIALIZE_ARGS.pReserved from
-		 * 'x-init-reserved' setting in the config. This only works with specific
-		 * PKCS#11 modules, and is non-standard use of that field.
-		 */
-		mod->init_args.pReserved = p11_dict_get (*config, "x-init-reserved");
 	}
 
+	/*
+	 * We support setting of CK_C_INITIALIZE_ARGS.pReserved from
+	 * 'x-init-reserved' setting in the config. This only works with specific
+	 * PKCS#11 modules, and is non-standard use of that field.
+	 */
+	mod->init_args.pReserved = p11_dict_get (*config, "x-init-reserved");
+
 	/* Take ownership of thes evariables */
 	p11_dict_free (mod->config);
 	mod->config = *config;
@@ -610,7 +612,7 @@ load_registered_modules_unlocked (void)
 }
 
 static CK_RV
-initialize_module_inlock_reentrant (Module *mod)
+initialize_module_inlock_reentrant (Module *mod, CK_C_INITIALIZE_ARGS *init_args)
 {
 	CK_RV rv = CKR_OK;
 	p11_thread_id_t self;
@@ -638,8 +640,12 @@ initialize_module_inlock_reentrant (Module *mod)
 	if (mod->initialize_called != p11_forkid) {
 		p11_debug ("C_Initialize: calling");
 
+		/* The init_args argument takes precedence over mod->init_args */
+		if (init_args == NULL)
+			init_args = &mod->init_args;
+
 		rv = mod->virt.funcs.C_Initialize (&mod->virt.funcs,
-		                                   &mod->init_args);
+						   init_args);
 
 		p11_debug ("C_Initialize: result: %lu", rv);
 
@@ -652,6 +658,10 @@ initialize_module_inlock_reentrant (Module *mod)
 		/* Module was already initialized, we don't call C_Finalize */
 		if (rv == CKR_CRYPTOKI_ALREADY_INITIALIZED)
 			rv = CKR_OK;
+
+		/* Matches the init count in finalize_module_inlock_reentrant() */
+		if (rv == CKR_OK)
+			mod->init_count = 0;
 	}
 
 	p11_mutex_unlock (&mod->initialize_mutex);
@@ -745,7 +755,7 @@ finalize_module_inlock_reentrant (Module *mod)
 		return CKR_OK;
 
 	/*
-	 * Becuase of the mutex unlock below, we temporarily increase
+	 * Because of the mutex unlock below, we temporarily increase
 	 * the ref count. This prevents module from being freed out
 	 * from ounder us.
 	 */
@@ -793,7 +803,7 @@ initialize_registered_inlock_reentrant (void)
 			if (mod->name == NULL || !is_module_enabled_unlocked (mod->name, mod->config))
 				continue;
 
-			rv = initialize_module_inlock_reentrant (mod);
+			rv = initialize_module_inlock_reentrant (mod, NULL);
 			if (rv != CKR_OK) {
 				if (mod->critical) {
 					p11_message ("initialization of critical module '%s' failed: %s",
@@ -1478,8 +1488,10 @@ managed_C_Initialize (CK_X_FUNCTION_LIST *self,
 		if (!sessions)
 			rv = CKR_HOST_MEMORY;
 		else
-			rv = initialize_module_inlock_reentrant (managed->mod);
+			rv = initialize_module_inlock_reentrant (managed->mod, init_args);
 		if (rv == CKR_OK) {
+			if (managed->sessions)
+				p11_dict_free (managed->sessions);
 			managed->sessions = sessions;
 			managed->initialized = p11_forkid;
 		} else {
@@ -1951,7 +1963,7 @@ p11_modules_load_inlock_reentrant (int flags,
  * If @flags contains the %P11_KIT_MODULE_CRITICAL flag then the
  * modules will all be treated as 'critical', regardless of the module
  * configuration. This means that a failure to load any module will
- * cause this funtion to fail.
+ * cause this function to fail.
  *
  * For unmanaged modules there is no guarantee to the state of the
  * modules. Other callers may be using the modules. Using unmanaged
@@ -2292,7 +2304,7 @@ p11_kit_initialize_module (CK_FUNCTION_LIST_PTR module)
 		if (rv == CKR_OK) {
 			mod = p11_dict_get (gl.unmanaged_by_funcs, module);
 			assert (mod != NULL);
-			rv = initialize_module_inlock_reentrant (mod);
+			rv = initialize_module_inlock_reentrant (mod, NULL);
 			if (rv != CKR_OK) {
 				p11_message ("module initialization failed: %s", p11_kit_strerror (rv));
 				p11_module_release_inlock_reentrant (module);
@@ -2326,17 +2338,15 @@ p11_module_load_inlock_reentrant (CK_FUNCTION_LIST *module,
 		}
 
 		/* If this was newly allocated, add it to the list */
-		if (rv == CKR_OK && allocated) {
+		if (allocated) {
 			if (!p11_dict_set (gl.modules, allocated, allocated) ||
 			    !p11_dict_set (gl.unmanaged_by_funcs, module, allocated))
 				return_val_if_reached (CKR_HOST_MEMORY);
 			allocated = NULL;
 		}
 
-		if (rv == CKR_OK) {
 			/* WARNING: Reentrancy can occur here */
 			rv = prepare_module_inlock_reentrant (mod, flags, result);
-		}
 
 		free (allocated);
 	}
@@ -2355,13 +2365,16 @@ p11_module_load_inlock_reentrant (CK_FUNCTION_LIST *module,
 
 /**
  * p11_kit_module_load:
- * @module_path: full file path of module library
+ * @module_path: relative or full file path of module library
  * @flags: flags to use when loading the module
  *
  * Load an arbitrary PKCS\#11 module from a dynamic library file, and
  * initialize it. Normally using the p11_kit_modules_load() function
  * is preferred.
  *
+ * A full file path or just (path/)filename relative to
+ * P11_MODULE_PATH are accepted.
+ *
  * Using this function to load modules allows coordination between multiple
  * callers of the same module in a single process. If @flags contains the
  * %P11_KIT_MODULE_UNMANAGED flag, then the modules will be not be loaded
@@ -2671,11 +2684,11 @@ p11_kit_load_initialize_module (const char *module_path,
 			if (rv == CKR_OK) {
 
 				/* WARNING: Reentrancy can occur here */
-				rv = initialize_module_inlock_reentrant (mod);
+				rv = initialize_module_inlock_reentrant (mod, NULL);
 			}
 		}
 
-		if (rv == CKR_OK && module) {
+		if (rv == CKR_OK) {
 			*module = unmanaged_for_module_inlock (mod);
 			assert (*module != NULL);
 		}
