@@ -41,6 +41,7 @@
 #include "library.h"
 #include "private.h"
 #include "message.h"
+#include "remote.h"
 #include "rpc.h"
 #include "rpc-message.h"
 
@@ -661,6 +662,9 @@ rpc_C_Initialize (CK_X_FUNCTION_LIST *self,
 	CK_C_INITIALIZE_ARGS init_args;
 	CK_BYTE_PTR handshake;
 	CK_ULONG n_handshake;
+	CK_BYTE reserved_present = 0;
+	CK_BYTE_PTR reserved = NULL;
+	CK_ULONG n_reserved;
 	CK_RV ret = CKR_OK;
 
 	p11_debug ("C_Initialize: enter");
@@ -677,6 +681,15 @@ rpc_C_Initialize (CK_X_FUNCTION_LIST *self,
 			p11_message ("invalid handshake received from connecting module");
 			ret = CKR_GENERAL_ERROR;
 		}
+	}
+
+	if (ret == CKR_OK) {
+		if (!p11_rpc_message_read_byte (msg, &reserved_present))
+			ret = PARSE_ERROR;
+	}
+
+	if (ret == CKR_OK) {
+		ret = proto_read_byte_array (msg, &reserved, &n_reserved);
 
 		assert (p11_rpc_message_is_verified (msg));
 	}
@@ -684,6 +697,7 @@ rpc_C_Initialize (CK_X_FUNCTION_LIST *self,
 	if (ret == CKR_OK) {
 		memset (&init_args, 0, sizeof (init_args));
 		init_args.flags = CKF_OS_LOCKING_OK;
+		init_args.pReserved = reserved_present ? reserved : NULL;
 
 		func = self->C_Initialize;
 		assert (func != NULL);
@@ -861,7 +875,7 @@ rpc_C_CloseAllSessions (CK_X_FUNCTION_LIST *self,
 {
 	CK_SLOT_ID slot_id;
 
-	/* Slot id becomes appartment so lower layers can tell clients apart. */
+	/* Slot id becomes apartment so lower layers can tell clients apart. */
 
 	BEGIN_CALL (CloseAllSessions);
 		IN_ULONG (slot_id);
@@ -1901,3 +1915,103 @@ p11_rpc_server_handle (CK_X_FUNCTION_LIST *self,
 	p11_rpc_message_clear (&msg);
 	return true;
 }
+
+int
+p11_kit_remote_serve_module (CK_FUNCTION_LIST *module,
+                             int in_fd,
+                             int out_fd)
+{
+	p11_rpc_status status;
+	unsigned char version;
+	p11_virtual virt;
+	p11_buffer options;
+	p11_buffer buffer;
+	size_t state;
+	int ret = 1;
+	int code;
+
+	return_val_if_fail (module != NULL, 1);
+
+	p11_buffer_init (&options, 0);
+	p11_buffer_init (&buffer, 0);
+
+	p11_virtual_init (&virt, &p11_virtual_base, module, NULL);
+
+	switch (read (in_fd, &version, 1)) {
+	case 0:
+		goto out;
+	case 1:
+		if (version != 0) {
+			p11_message ("unsupported version received: %d", (int)version);
+			goto out;
+		}
+		break;
+	default:
+		p11_message_err (errno, "couldn't read credential byte");
+		goto out;
+	}
+
+	version = 0;
+	switch (write (out_fd, &version, out_fd)) {
+	case 1:
+		break;
+	default:
+		p11_message_err (errno, "couldn't write credential byte");
+		goto out;
+	}
+
+	status = P11_RPC_OK;
+	while (status == P11_RPC_OK) {
+		state = 0;
+		code = 0;
+
+		do {
+			status = p11_rpc_transport_read (in_fd, &state, &code,
+			                                 &options, &buffer);
+		} while (status == P11_RPC_AGAIN);
+
+		switch (status) {
+		case P11_RPC_OK:
+			break;
+		case P11_RPC_EOF:
+			ret = 0;
+			continue;
+		case P11_RPC_AGAIN:
+			assert_not_reached ();
+		case P11_RPC_ERROR:
+			p11_message_err (errno, "failed to read rpc message");
+			goto out;
+		}
+
+		if (!p11_rpc_server_handle (&virt.funcs, &buffer, &buffer)) {
+			p11_message ("unexpected error handling rpc message");
+			goto out;
+		}
+
+		state = 0;
+		options.len = 0;
+		do {
+			status = p11_rpc_transport_write (out_fd, &state, code,
+			                                  &options, &buffer);
+		} while (status == P11_RPC_AGAIN);
+
+		switch (status) {
+		case P11_RPC_OK:
+			break;
+		case P11_RPC_EOF:
+		case P11_RPC_AGAIN:
+			assert_not_reached ();
+		case P11_RPC_ERROR:
+			p11_message_err (errno, "failed to write rpc message");
+			goto out;
+		}
+	}
+
+out:
+	p11_buffer_uninit (&buffer);
+	p11_buffer_uninit (&options);
+
+	p11_virtual_uninit (&virt);
+
+	return ret;
+}
