@@ -38,7 +38,7 @@
 #include "buffer.h"
 #define P11_DEBUG_FLAG P11_DEBUG_URI
 #include "debug.h"
-#include "library.h"
+#include "message.h"
 #include "pkcs11.h"
 #include "private.h"
 #include "p11-kit.h"
@@ -62,7 +62,7 @@
  * <code><literallayout>
  *      pkcs11:token=The\%20Software\%20PKCS\#11\%20softtoken;
  *          manufacturer=Snake\%20Oil,\%20Inc.;serial=;object=my-certificate;
- *          model=1.0;object-type=cert;id=\%69\%95\%3e\%5c\%f4\%bd\%ec\%91
+ *          model=1.0;type=cert;id=\%69\%95\%3e\%5c\%f4\%bd\%ec\%91
  * </literallayout></code>
  *
  * You can use p11_kit_uri_parse() to parse such a URI, and p11_kit_uri_format()
@@ -90,6 +90,7 @@
  * P11KitUriType:
  * @P11_KIT_URI_FOR_OBJECT: The URI represents one or more objects
  * @P11_KIT_URI_FOR_TOKEN: The URI represents one or more tokens
+ * @P11_KIT_URI_FOR_SLOT: The URI represents one or more slots
  * @P11_KIT_URI_FOR_MODULE: The URI represents one or more modules
  * @P11_KIT_URI_FOR_MODULE_WITH_VERSION: The URI represents a module with
  *     a specific version.
@@ -141,17 +142,18 @@
 struct p11_kit_uri {
 	bool unrecognized;
 	CK_INFO module;
+	CK_SLOT_INFO slot;
 	CK_TOKEN_INFO token;
 	CK_ATTRIBUTE *attrs;
 	char *pin_source;
+	char *pin_value;
+	CK_SLOT_ID slot_id;
 };
 
-const static char WHITESPACE[] = " \n\r\v";
-
 static char *
-key_decode (const char *value, const char *end)
+strip_whitespace (const char *value)
 {
-	size_t length = (end - value);
+	size_t length = strlen (value);
 	char *at, *pos;
 	char *key;
 
@@ -162,9 +164,9 @@ key_decode (const char *value, const char *end)
 	key[length] = '\0';
 
 	/* Do we have any whitespace? Strip it out. */
-	if (strcspn (key, WHITESPACE) != length) {
+	if (strcspn (key, P11_URL_WHITESPACE) != length) {
 		for (at = key, pos = key; pos != key + length + 1; ++pos) {
-			if (!strchr (WHITESPACE, *pos))
+			if (!strchr (P11_URL_WHITESPACE, *pos))
 				*(at++) = *pos;
 		}
 		*at = '\0';
@@ -261,6 +263,98 @@ p11_kit_uri_match_module_info (P11KitUri *uri, CK_INFO_PTR info)
 	return p11_match_uri_module_info (&uri->module, info);
 }
 
+/**
+ * p11_kit_uri_get_slot_info:
+ * @uri: the URI
+ *
+ * Get the <code>CK_SLOT_INFO</code> structure associated with this URI.
+ *
+ * If this is a parsed URI, then the fields corresponding to slot parts of
+ * the URI will be filled in. Any slot URI parts that were missing will have
+ * their fields filled with zeros.
+ *
+ * If the caller wishes to setup information for building a URI, then relevant
+ * fields should be filled in. Fields that should not appear as parts in the
+ * resulting URI should be filled with zeros.
+ *
+ * Returns: A pointer to the <code>CK_INFO</code> structure.
+ */
+CK_SLOT_INFO_PTR
+p11_kit_uri_get_slot_info (P11KitUri *uri)
+{
+	return_val_if_fail (uri != NULL, NULL);
+	return &uri->slot;
+}
+
+int
+p11_match_uri_slot_info (CK_SLOT_INFO_PTR one,
+			 CK_SLOT_INFO_PTR two)
+{
+	return (match_struct_string (one->slotDescription,
+				     two->slotDescription,
+				     sizeof (one->slotDescription)) &&
+		match_struct_string (one->manufacturerID,
+				     two->manufacturerID,
+				     sizeof (one->manufacturerID)));
+}
+
+/**
+ * p11_kit_uri_match_slot_info:
+ * @uri: the URI
+ * @slot_info: the structure to match against the URI
+ *
+ * Match a <code>CK_SLOT_INFO</code> structure against the slot parts of this
+ * URI.
+ *
+ * Only the fields of the <code>CK_SLOT_INFO</code> structure that are valid
+ * for use in a URI will be matched. A URI part that was not specified in the
+ * URI will match any value in the structure. If during the URI parsing any
+ * unrecognized parts were encountered then this match will fail.
+ *
+ * Returns: 1 if the URI matches, 0 if not.
+ */
+int
+p11_kit_uri_match_slot_info (P11KitUri *uri, CK_SLOT_INFO_PTR slot_info)
+{
+	return_val_if_fail (uri != NULL, 0);
+	return_val_if_fail (slot_info != NULL, 0);
+
+	if (uri->unrecognized)
+		return 0;
+
+	return p11_match_uri_slot_info (&uri->slot, slot_info);
+}
+
+/**
+ * p11_kit_uri_get_slot_id:
+ * @uri: The URI
+ *
+ * Get the 'slot-id' part of the URI.
+ *
+ * Returns: The slot-id or <code>(CK_SLOT_ID)-1</code> if not set.
+ */
+CK_SLOT_ID
+p11_kit_uri_get_slot_id (P11KitUri *uri)
+{
+	return_val_if_fail (uri != NULL, (CK_SLOT_ID)-1);
+	return uri->slot_id;
+}
+
+/**
+ * p11_kit_uri_set_slot_id:
+ * @uri: The URI
+ * @slot_id: The new slot-id
+ *
+ * Set the 'slot-id' part of the URI.
+ */
+void
+p11_kit_uri_set_slot_id (P11KitUri  *uri,
+			 CK_SLOT_ID  slot_id)
+{
+	return_if_fail (uri != NULL);
+	uri->slot_id = slot_id;
+}
+
 /**
  * p11_kit_uri_get_token_info:
  * @uri: the URI
@@ -415,14 +509,14 @@ p11_kit_uri_clear_attribute (P11KitUri *uri, CK_ATTRIBUTE_TYPE attr_type)
 CK_ATTRIBUTE_PTR
 p11_kit_uri_get_attributes (P11KitUri *uri, CK_ULONG_PTR n_attrs)
 {
-	static CK_ATTRIBUTE empty = { CKA_INVALID, NULL, 0UL };
+	static const CK_ATTRIBUTE terminator = { CKA_INVALID, NULL, 0UL };
 
 	return_val_if_fail (uri != NULL, NULL);
 
 	if (!uri->attrs) {
 		if (n_attrs)
 			*n_attrs = 0;
-		return &empty;
+		return (CK_ATTRIBUTE_PTR)&terminator;
 	}
 
 	if (n_attrs)
@@ -541,6 +635,39 @@ p11_kit_uri_any_unrecognized (P11KitUri *uri)
 	return uri->unrecognized;
 }
 
+/**
+ * p11_kit_uri_get_pin_value:
+ * @uri: The URI
+ *
+ * Get the 'pin-value' part of the URI. This is used by some applications to
+ * read the PIN for logging into a PKCS\#11 token.
+ *
+ * Returns: The pin-value or %NULL if not present.
+ */
+const char*
+p11_kit_uri_get_pin_value (P11KitUri *uri)
+{
+	return_val_if_fail (uri != NULL, NULL);
+	return uri->pin_value;
+}
+
+/**
+ * p11_kit_uri_set_pin_value:
+ * @uri: The URI
+ * @pin: The new pin-value
+ *
+ * Set the 'pin-value' part of the URI. This is used by some applications to
+ * specify the PIN for logging into a PKCS\#11 token.
+ */
+void
+p11_kit_uri_set_pin_value (P11KitUri *uri, const char *pin)
+{
+	return_if_fail (uri != NULL);
+	free (uri->pin_value);
+	uri->pin_value = pin ? strdup (pin) : NULL;
+}
+
+
 /**
  * p11_kit_uri_get_pin_source:
  * @uri: The URI
@@ -621,10 +748,23 @@ p11_kit_uri_new (void)
 	/* So that it matches anything */
 	uri->module.libraryVersion.major = (CK_BYTE)-1;
 	uri->module.libraryVersion.minor = (CK_BYTE)-1;
+	uri->slot_id = (CK_SLOT_ID)-1;
 
 	return uri;
 }
 
+static void
+format_name_equals (p11_buffer *buffer,
+                    bool *is_first,
+                    const char *name)
+{
+	if (!*is_first)
+		p11_buffer_add (buffer, ";", 1);
+	p11_buffer_add (buffer, name, -1);
+	p11_buffer_add (buffer, "=", 1);
+	*is_first = false;
+}
+
 static bool
 format_raw_string (p11_buffer *buffer,
                    bool *is_first,
@@ -635,12 +775,8 @@ format_raw_string (p11_buffer *buffer,
 	if (!value)
 		return true;
 
-	if (!*is_first)
-		p11_buffer_add (buffer, ";", 1);
-	p11_buffer_add (buffer, name, -1);
-	p11_buffer_add (buffer, "=", 1);
+	format_name_equals (buffer, is_first, name);
 	p11_buffer_add (buffer, value, -1);
-	*is_first = false;
 
 	return p11_buffer_ok (buffer);
 }
@@ -653,16 +789,14 @@ format_encode_string (p11_buffer *buffer,
                       size_t n_value,
                       bool force)
 {
-	char *encoded;
-	bool ret;
+	/* Not set */
+	if (!value)
+		return true;
 
-	encoded = p11_url_encode (value, value + n_value,
-	                          force ? "" : P11_URL_VERBATIM, NULL);
-	return_val_if_fail (encoded != NULL, false);
+	format_name_equals (buffer, is_first, name);
+	p11_url_encode (value, value + n_value, force ? "" : P11_URL_VERBATIM, buffer);
 
-	ret = format_raw_string (buffer, is_first, name, encoded);
-	free (encoded);
-	return ret;
+	return p11_buffer_ok (buffer);
 }
 
 
@@ -753,6 +887,22 @@ format_struct_version (p11_buffer *buffer,
 	return format_raw_string (buffer, is_first, name, buf);
 }
 
+static bool
+format_ulong (p11_buffer *buffer,
+	      bool *is_first,
+	      const char *name,
+	      CK_ULONG value)
+{
+	char buf[64];
+
+	/* Not set */
+	if (value == (CK_ULONG)-1)
+		return true;
+
+	snprintf (buf, sizeof (buf), "%lu", value);
+	return format_raw_string (buffer, is_first, name, buf);
+}
+
 /**
  * p11_kit_uri_format:
  * @uri: The URI.
@@ -814,6 +964,19 @@ p11_kit_uri_format (P11KitUri *uri, P11KitUriType uri_type, char **string)
 		}
 	}
 
+	if ((uri_type & P11_KIT_URI_FOR_SLOT) == P11_KIT_URI_FOR_SLOT) {
+		if (!format_struct_string (&buffer, &is_first, "slot-description",
+		                           uri->slot.slotDescription,
+		                           sizeof (uri->slot.slotDescription)) ||
+		    !format_struct_string (&buffer, &is_first, "slot-manufacturer",
+		                           uri->slot.manufacturerID,
+		                           sizeof (uri->slot.manufacturerID)) ||
+		    !format_ulong (&buffer, &is_first, "slot-id",
+				   uri->slot_id)) {
+			return_val_if_reached (P11_KIT_URI_UNEXPECTED);
+		}
+	}
+
 	if ((uri_type & P11_KIT_URI_FOR_TOKEN) == P11_KIT_URI_FOR_TOKEN) {
 		if (!format_struct_string (&buffer, &is_first, "model",
 		                           uri->token.model,
@@ -841,7 +1004,7 @@ p11_kit_uri_format (P11KitUri *uri, P11KitUriType uri_type, char **string)
 			return_val_if_reached (P11_KIT_URI_UNEXPECTED);
 		}
 
-		if (!format_attribute_class (&buffer, &is_first, "object-type",
+		if (!format_attribute_class (&buffer, &is_first, "type",
 		                             p11_kit_uri_get_attribute (uri, CKA_CLASS))) {
 			return_val_if_reached (P11_KIT_URI_UNEXPECTED);
 		}
@@ -855,24 +1018,41 @@ p11_kit_uri_format (P11KitUri *uri, P11KitUriType uri_type, char **string)
 		}
 	}
 
+	if (uri->pin_value) {
+		if (!format_encode_string (&buffer, &is_first, "pin-value",
+		                           (const unsigned char*)uri->pin_value,
+		                           strlen (uri->pin_value), 0)) {
+			return_val_if_reached (P11_KIT_URI_UNEXPECTED);
+		}
+	}
+
 	return_val_if_fail (p11_buffer_ok (&buffer), P11_KIT_URI_UNEXPECTED);
 	*string = p11_buffer_steal (&buffer, NULL);
 	return P11_KIT_URI_OK;
 }
 
+static bool
+str_range_equal (const char *input, const char *start, const char *end)
+{
+	return strlen (input) == end - start &&
+		memcmp (input, start, end - start) == 0;
+}
+
 static int
-parse_string_attribute (const char *name, const char *start, const char *end,
-                        P11KitUri *uri)
+parse_string_attribute (const char *name_start, const char *name_end,
+			const char *start, const char *end,
+			P11KitUri *uri)
 {
 	unsigned char *value;
 	CK_ATTRIBUTE_TYPE type;
 	size_t length;
 
+	assert (name_start <= name_end);
 	assert (start <= end);
 
-	if (strcmp ("id", name) == 0)
+	if (str_range_equal ("id", name_start, name_end))
 		type = CKA_ID;
-	else if (strcmp ("object", name) == 0)
+	else if (str_range_equal ("object", name_start, name_end))
 		type = CKA_LABEL;
 	else
 		return 0;
@@ -886,42 +1066,38 @@ parse_string_attribute (const char *name, const char *start, const char *end,
 }
 
 static int
-parse_class_attribute (const char *name, const char *start, const char *end,
-                       P11KitUri *uri)
+parse_class_attribute (const char *name_start, const char *name_end,
+		       const char *start, const char *end,
+		       P11KitUri *uri)
 {
 	CK_OBJECT_CLASS klass = 0;
 	CK_ATTRIBUTE attr;
-	char *value;
 
+	assert (name_start <= name_end);
 	assert (start <= end);
 
-	if (strcmp ("objecttype", name) != 0 &&
-	    strcmp ("object-type", name) != 0)
+	if (!str_range_equal ("objecttype", name_start, name_end) &&
+	    !str_range_equal ("object-type", name_start, name_end) &&
+	    !str_range_equal ("type", name_start, name_end))
 		return 0;
 
-	value = key_decode (start, end);
-	return_val_if_fail (value != NULL, P11_KIT_URI_UNEXPECTED);
-
-	if (strcmp (value, "cert") == 0)
+	if (str_range_equal ("cert", start, end))
 		klass = CKO_CERTIFICATE;
-	else if (strcmp (value, "public") == 0)
+	else if (str_range_equal ("public", start, end))
 		klass = CKO_PUBLIC_KEY;
-	else if (strcmp (value, "private") == 0)
+	else if (str_range_equal ("private", start, end))
 		klass = CKO_PRIVATE_KEY;
-	else if (strcmp (value, "secretkey") == 0)
+	else if (str_range_equal ("secretkey", start, end))
 		klass = CKO_SECRET_KEY;
-	else if (strcmp (value, "secret-key") == 0)
+	else if (str_range_equal ("secret-key", start, end))
 		klass = CKO_SECRET_KEY;
-	else if (strcmp (value, "data") == 0)
+	else if (str_range_equal ("data", start, end))
 		klass = CKO_DATA;
 	else {
-		free (value);
 		uri->unrecognized = true;
 		return 1;
 	}
 
-	free (value);
-
 	attr.pValue = &klass;
 	attr.ulValueLen = sizeof (klass);
 	attr.type = CKA_CLASS;
@@ -958,24 +1134,26 @@ parse_struct_info (unsigned char *where, size_t length, const char *start,
 }
 
 static int
-parse_token_info (const char *name, const char *start, const char *end,
-                  P11KitUri *uri)
+parse_token_info (const char *name_start, const char *name_end,
+		  const char *start, const char *end,
+		  P11KitUri *uri)
 {
 	unsigned char *where;
 	size_t length;
 
+	assert (name_start <= name_end);
 	assert (start <= end);
 
-	if (strcmp (name, "model") == 0) {
+	if (str_range_equal ("model", name_start, name_end)) {
 		where = uri->token.model;
 		length = sizeof (uri->token.model);
-	} else if (strcmp (name, "manufacturer") == 0) {
+	} else if (str_range_equal ("manufacturer", name_start, name_end)) {
 		where = uri->token.manufacturerID;
 		length = sizeof (uri->token.manufacturerID);
-	} else if (strcmp (name, "serial") == 0) {
+	} else if (str_range_equal ("serial", name_start, name_end)) {
 		where = uri->token.serialNumber;
 		length = sizeof (uri->token.serialNumber);
-	} else if (strcmp (name, "token") == 0) {
+	} else if (str_range_equal ("token", name_start, name_end)) {
 		where = uri->token.label;
 		length = sizeof (uri->token.label);
 	} else {
@@ -985,15 +1163,11 @@ parse_token_info (const char *name, const char *start, const char *end,
 	return parse_struct_info (where, length, start, end, uri);
 }
 
-static int
+static long
 atoin (const char *start, const char *end)
 {
-	int ret = 0;
+	long ret = 0;
 	while (start != end) {
-		if (strchr (WHITESPACE, *start)) {
-			start++;
-			continue;
-		}
 		if (*start < '0' || *start > '9')
 			return -1;
 		ret *= 10;
@@ -1036,12 +1210,57 @@ parse_struct_version (const char *start, const char *end, CK_VERSION_PTR version
 }
 
 static int
-parse_module_version_info (const char *name, const char *start, const char *end,
-                           P11KitUri *uri)
+parse_slot_info (const char *name_start, const char *name_end,
+                 const char *start, const char *end,
+                 P11KitUri *uri)
 {
+	unsigned char *where;
+	size_t length;
+
+	assert (name_start <= name_end);
 	assert (start <= end);
 
-	if (strcmp (name, "library-version") == 0)
+	if (str_range_equal ("slot-description", name_start, name_end)) {
+		where = uri->slot.slotDescription;
+		length = sizeof (uri->slot.slotDescription);
+	} else if (str_range_equal ("slot-manufacturer", name_start, name_end)) {
+		where = uri->slot.manufacturerID;
+		length = sizeof (uri->slot.manufacturerID);
+	} else {
+		return 0;
+	}
+
+	return parse_struct_info (where, length, start, end, uri);
+}
+
+static int
+parse_slot_id (const char *name_start, const char *name_end,
+	       const char *start, const char *end,
+	       P11KitUri *uri)
+{
+	assert (name_start <= name_end);
+	assert (start <= end);
+
+	if (str_range_equal ("slot-id", name_start, name_end)) {
+		long val;
+		val = atoin (start, end);
+		if (val < 0)
+			return P11_KIT_URI_BAD_SYNTAX;
+		uri->slot_id = (CK_SLOT_ID)val;
+		return 1;
+	}
+	return 0;
+}
+
+static int
+parse_module_version_info (const char *name_start, const char *name_end,
+			   const char *start, const char *end,
+			   P11KitUri *uri)
+{
+	assert (name_start <= name_end);
+	assert (start <= end);
+
+	if (str_range_equal ("library-version", name_start, name_end))
 		return parse_struct_version (start, end,
 		                             &uri->module.libraryVersion);
 
@@ -1049,18 +1268,20 @@ parse_module_version_info (const char *name, const char *start, const char *end,
 }
 
 static int
-parse_module_info (const char *name, const char *start, const char *end,
-                   P11KitUri *uri)
+parse_module_info (const char *name_start, const char *name_end,
+		   const char *start, const char *end,
+		   P11KitUri *uri)
 {
 	unsigned char *where;
 	size_t length;
 
+	assert (name_start <= name_end);
 	assert (start <= end);
 
-	if (strcmp (name, "library-description") == 0) {
+	if (str_range_equal ("library-description", name_start, name_end)) {
 		where = uri->module.libraryDescription;
 		length = sizeof (uri->module.libraryDescription);
-	} else if (strcmp (name, "library-manufacturer") == 0) {
+	} else if (str_range_equal ("library-manufacturer", name_start, name_end)) {
 		where = uri->module.manufacturerID;
 		length = sizeof (uri->module.manufacturerID);
 	} else {
@@ -1071,21 +1292,30 @@ parse_module_info (const char *name, const char *start, const char *end,
 }
 
 static int
-parse_extra_info (const char *name, const char *start, const char *end,
-                  P11KitUri *uri)
+parse_extra_info (const char *name_start, const char *name_end,
+		  const char *start, const char *end,
+		  P11KitUri *uri)
 {
 	unsigned char *pin_source;
 
+	assert (name_start <= name_end);
 	assert (start <= end);
 
-	if (strcmp (name, "pinfile") == 0 ||
-	    strcmp (name, "pin-source") == 0) {
+	if (str_range_equal ("pinfile", name_start, name_end) ||
+	    str_range_equal ("pin-source", name_start, name_end)) {
 		pin_source = p11_url_decode (start, end, P11_URL_WHITESPACE, NULL);
 		if (pin_source == NULL)
 			return P11_KIT_URI_BAD_ENCODING;
 		free (uri->pin_source);
 		uri->pin_source = (char*)pin_source;
 		return 1;
+	} else if (str_range_equal ("pin-value", name_start, name_end)) {
+		pin_source = p11_url_decode (start, end, P11_URL_WHITESPACE, NULL);
+		if (pin_source == NULL)
+			return P11_KIT_URI_BAD_ENCODING;
+		free (uri->pin_value);
+		uri->pin_value = (char*)pin_source;
+		return 1;
 	}
 
 	return 0;
@@ -1120,21 +1350,32 @@ p11_kit_uri_parse (const char *string, P11KitUriType uri_type,
                    P11KitUri *uri)
 {
 	const char *spos, *epos;
-	char *key = NULL;
 	int ret;
+	size_t length;
+	char *allocated = NULL;
 
 	assert (string);
 	assert (uri);
 
-	epos = strchr (string, ':');
-	if (epos == NULL)
-		return P11_KIT_URI_BAD_SCHEME;
-	key = key_decode (string, epos);
-	ret = strcmp (key, P11_KIT_URI_SCHEME);
-	free (key);
+	/* If STRING contains any whitespace, create a copy of the
+	 * string and strip it out */
+	length = strcspn (string, P11_URL_WHITESPACE);
+	if (strspn (string + length, P11_URL_WHITESPACE) > 0) {
+		allocated = strip_whitespace (string);
+		return_val_if_fail (allocated != NULL, P11_KIT_URI_UNEXPECTED);
+		string = allocated;
+	}
 
-	if (ret != 0)
+	epos = strchr (string, ':');
+	if (epos == NULL) {
+		free (allocated);
 		return P11_KIT_URI_BAD_SCHEME;
+	}
+	ret = strncmp (string, P11_KIT_URI_SCHEME, strlen (P11_KIT_URI_SCHEME));
+	if (ret != 0) {
+		free (allocated);
+		return P11_KIT_URI_BAD_SCHEME;
+	}
 
 	string = epos + 1;
 
@@ -1148,6 +1389,9 @@ p11_kit_uri_parse (const char *string, P11KitUriType uri_type,
 	uri->unrecognized = 0;
 	free (uri->pin_source);
 	uri->pin_source = NULL;
+	free (uri->pin_value);
+	uri->pin_value = NULL;
+	uri->slot_id = (CK_SLOT_ID)-1;
 
 	for (;;) {
 		spos = strchr (string, ';');
@@ -1159,30 +1403,33 @@ p11_kit_uri_parse (const char *string, P11KitUriType uri_type,
 		}
 
 		epos = strchr (string, '=');
-		if (epos == NULL || spos == string || epos == string || epos >= spos)
+		if (epos == NULL || spos == string || epos == string || epos >= spos) {
+			free (allocated);
 			return P11_KIT_URI_BAD_SYNTAX;
-
-		key = key_decode (string, epos);
-		return_val_if_fail (key != NULL, P11_KIT_URI_UNEXPECTED);
-		epos++;
+		}
 
 		ret = 0;
 		if ((uri_type & P11_KIT_URI_FOR_OBJECT) == P11_KIT_URI_FOR_OBJECT)
-			ret = parse_string_attribute (key, epos, spos, uri);
+			ret = parse_string_attribute (string, epos, epos + 1, spos, uri);
 		if (ret == 0 && (uri_type & P11_KIT_URI_FOR_OBJECT) == P11_KIT_URI_FOR_OBJECT)
-			ret = parse_class_attribute (key, epos, spos, uri);
+			ret = parse_class_attribute (string, epos, epos + 1, spos, uri);
 		if (ret == 0 && (uri_type & P11_KIT_URI_FOR_TOKEN) == P11_KIT_URI_FOR_TOKEN)
-			ret = parse_token_info (key, epos, spos, uri);
+			ret = parse_token_info (string, epos, epos + 1, spos, uri);
+		if (ret == 0 && (uri_type & P11_KIT_URI_FOR_SLOT) == P11_KIT_URI_FOR_SLOT)
+			ret = parse_slot_info (string, epos, epos + 1, spos, uri);
+		if (ret == 0 && (uri_type & P11_KIT_URI_FOR_SLOT) == P11_KIT_URI_FOR_SLOT)
+			ret = parse_slot_id (string, epos, epos + 1, spos, uri);
 		if (ret == 0 && (uri_type & P11_KIT_URI_FOR_MODULE) == P11_KIT_URI_FOR_MODULE)
-			ret = parse_module_info (key, epos, spos, uri);
+			ret = parse_module_info (string, epos, epos + 1, spos, uri);
 		if (ret == 0 && (uri_type & P11_KIT_URI_FOR_MODULE_WITH_VERSION) == P11_KIT_URI_FOR_MODULE_WITH_VERSION)
-			ret = parse_module_version_info (key, epos, spos, uri);
+			ret = parse_module_version_info (string, epos, epos + 1, spos, uri);
 		if (ret == 0)
-			ret = parse_extra_info (key, epos, spos, uri);
-		free (key);
+			ret = parse_extra_info (string, epos, epos + 1, spos, uri);
 
-		if (ret < 0)
+		if (ret < 0) {
+			free (allocated);
 			return ret;
+		}
 		if (ret == 0)
 			uri->unrecognized = true;
 
@@ -1191,6 +1438,7 @@ p11_kit_uri_parse (const char *string, P11KitUriType uri_type,
 		string = spos + 1;
 	}
 
+	free (allocated);
 	return P11_KIT_URI_OK;
 }
 
@@ -1208,6 +1456,7 @@ p11_kit_uri_free (P11KitUri *uri)
 
 	p11_attrs_free (uri->attrs);
 	free (uri->pin_source);
+	free (uri->pin_value);
 	free (uri);
 }
 
